@@ -24,6 +24,7 @@
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
+#undef HAVE_CONFIG_H  // Workaround for Fribidi 1.0.5 and earlier
 #endif
 
 #include <assert.h>
@@ -191,6 +192,7 @@ struct _raqm {
   _raqm_flags_t    flags;
 
   int              ft_loadflags;
+  int              invisible_glyph;
 };
 
 struct _raqm_run {
@@ -306,6 +308,7 @@ raqm_create (void)
   rq->flags = RAQM_FLAG_NONE;
 
   rq->ft_loadflags = -1;
+  rq->invisible_glyph = 0;
 
   return rq;
 }
@@ -766,6 +769,53 @@ raqm_set_freetype_load_flags (raqm_t *rq,
   return true;
 }
 
+/**
+ * raqm_set_invisible_glyph:
+ * @rq: a #raqm_t.
+ * @gid: glyph id to use for invisible glyphs.
+ *
+ * Sets the glyph id to be used for invisible glyhphs.
+ *
+ * If @gid is negative, invisible glyphs will be suppressed from the output.
+ * This requires HarfBuzz 1.8.0 or later. If raqm is used with an earlier
+ * HarfBuzz version, the return value will be %false and the shaping behavior
+ * does not change.
+ *
+ * If @gid is zero, invisible glyphs will be rendered as space.
+ * This works on all versions of HarfBuzz.
+ *
+ * If @gid is a positive number, it will be used for invisible glyphs.
+ * This requires a version of HarfBuzz that has
+ * hb_buffer_set_invisible_glyph(). For older versions, the return value
+ * will be %false and the shaping behavior does not change.
+ *
+ * Return value:
+ * %true if no errors happened, %false otherwise.
+ *
+ * Since: 0.6
+ */
+bool
+raqm_set_invisible_glyph (raqm_t *rq,
+                          int gid)
+{
+  if (!rq)
+    return false;
+
+#ifndef HAVE_HB_BUFFER_SET_INVISIBLE_GLYPH
+  if (gid > 0)
+    return false;
+#endif
+
+#if !defined(HAVE_DECL_HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES) || \
+    !HAVE_DECL_HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
+  if (gid < 0)
+    return false;
+#endif
+
+  rq->invisible_glyph = gid;
+  return true;
+}
+
 static bool
 _raqm_itemize (raqm_t *rq);
 
@@ -1093,7 +1143,7 @@ _raqm_itemize (raqm_t *rq)
   if (rq->base_dir == RAQM_DIRECTION_TTB)
   {
     /* Treat every thing as LTR in vertical text */
-    max_level = 0;
+    max_level = 1;
     memset (types, FRIBIDI_TYPE_LTR, rq->text_len);
     memset (levels, 0, rq->text_len);
     rq->resolved_dir = RAQM_DIRECTION_LTR;
@@ -1266,7 +1316,7 @@ typedef struct {
   size_t       capacity;
   size_t       size;
   int         *pair_index;
-  hb_script_t *scripts;
+  hb_script_t *script;
 } _raqm_stack_t;
 
 /* Special paired characters for script detection */
@@ -1295,7 +1345,7 @@ static const FriBidiChar paired_chars[] =
 static void
 _raqm_stack_free (_raqm_stack_t *stack)
 {
-  free (stack->scripts);
+  free (stack->script);
   free (stack->pair_index);
   free (stack);
 }
@@ -1309,8 +1359,8 @@ _raqm_stack_new (size_t max)
   if (!stack)
     return NULL;
 
-  stack->scripts = malloc (sizeof (hb_script_t) * max);
-  if (!stack->scripts)
+  stack->script = malloc (sizeof (hb_script_t) * max);
+  if (!stack->script)
   {
     _raqm_stack_free (stack);
     return NULL;
@@ -1352,13 +1402,13 @@ _raqm_stack_top (_raqm_stack_t *stack)
     return HB_SCRIPT_INVALID; /* XXX: check this */
   }
 
-  return stack->scripts[stack->size];
+  return stack->script[stack->size];
 }
 
 static bool
-_raqm_stack_push (_raqm_stack_t      *stack,
-            hb_script_t script,
-            int         pi)
+_raqm_stack_push (_raqm_stack_t *stack,
+                  hb_script_t    script,
+                  int            pair_index)
 {
   if (stack->size == stack->capacity)
   {
@@ -1367,14 +1417,14 @@ _raqm_stack_push (_raqm_stack_t      *stack,
   }
 
   stack->size++;
-  stack->scripts[stack->size] = script;
-  stack->pair_index[stack->size] = pi;
+  stack->script[stack->size] = script;
+  stack->pair_index[stack->size] = pair_index;
 
   return true;
 }
 
 static int
-get_pair_index (const FriBidiChar ch)
+_get_pair_index (const FriBidiChar ch)
 {
   int lower = 0;
   int upper = paired_len - 1;
@@ -1394,7 +1444,6 @@ get_pair_index (const FriBidiChar ch)
 }
 
 #define STACK_IS_EMPTY(script)     ((script)->size <= 0)
-#define STACK_IS_NOT_EMPTY(script) (! STACK_IS_EMPTY(script))
 #define IS_OPEN(pair_index)        (((pair_index) & 1) == 0)
 
 /* Resolve the script for each character in the input string, if the character
@@ -1407,12 +1456,12 @@ _raqm_resolve_scripts (raqm_t *rq)
 {
   int last_script_index = -1;
   int last_set_index = -1;
-  hb_script_t last_script_value = HB_SCRIPT_INVALID;
+  hb_script_t last_script = HB_SCRIPT_INVALID;
   _raqm_stack_t *stack = NULL;
+  hb_unicode_funcs_t* unicode_funcs = hb_unicode_funcs_get_default ();
 
   for (size_t i = 0; i < rq->text_len; ++i)
-    rq->text_info[i].script = hb_unicode_script (
-        hb_unicode_funcs_get_default (), rq->text[i]);
+    rq->text_info[i].script = hb_unicode_script (unicode_funcs, rq->text[i]);
 
 #ifdef RAQM_TESTING
   RAQM_TEST ("Before script detection:\n");
@@ -1430,16 +1479,15 @@ _raqm_resolve_scripts (raqm_t *rq)
 
   for (int i = 0; i < (int) rq->text_len; i++)
   {
-    if (rq->text_info[i].script == HB_SCRIPT_COMMON &&
-        last_script_index != -1)
+    if (rq->text_info[i].script == HB_SCRIPT_COMMON && last_script_index != -1)
     {
-      int pair_index = get_pair_index (rq->text[i]);
+      int pair_index = _get_pair_index (rq->text[i]);
       if (pair_index >= 0)
       {
         if (IS_OPEN (pair_index))
         {
           /* is a paired character */
-          rq->text_info[i].script = last_script_value;
+          rq->text_info[i].script = last_script;
           last_set_index = i;
           _raqm_stack_push (stack, rq->text_info[i].script, pair_index);
         }
@@ -1448,47 +1496,57 @@ _raqm_resolve_scripts (raqm_t *rq)
           /* is a close paired character */
           /* find matching opening (by getting the last even index for current
            * odd index) */
-          int pi = pair_index & ~1;
-          while (STACK_IS_NOT_EMPTY (stack) &&
-                 stack->pair_index[stack->size] != pi)
+          while (!STACK_IS_EMPTY (stack) &&
+                 stack->pair_index[stack->size] != (pair_index & ~1))
           {
             _raqm_stack_pop (stack);
           }
-          if (STACK_IS_NOT_EMPTY (stack))
+          if (!STACK_IS_EMPTY (stack))
           {
             rq->text_info[i].script = _raqm_stack_top (stack);
-            last_script_value = rq->text_info[i].script;
+            last_script = rq->text_info[i].script;
             last_set_index = i;
           }
           else
           {
-            rq->text_info[i].script = last_script_value;
+            rq->text_info[i].script = last_script;
             last_set_index = i;
           }
         }
       }
       else
       {
-        rq->text_info[i].script = last_script_value;
+        rq->text_info[i].script = last_script;
         last_set_index = i;
       }
     }
     else if (rq->text_info[i].script == HB_SCRIPT_INHERITED &&
              last_script_index != -1)
     {
-      rq->text_info[i].script = last_script_value;
+      rq->text_info[i].script = last_script;
       last_set_index = i;
     }
     else
     {
       for (int j = last_set_index + 1; j < i; ++j)
         rq->text_info[j].script = rq->text_info[i].script;
-      last_script_value = rq->text_info[i].script;
+      last_script = rq->text_info[i].script;
       last_script_index = i;
       last_set_index = i;
     }
   }
 
+  /* Loop backwards and change any remaining Common or Inherit characters to
+   * take the script if the next character.
+   * https://github.com/HOST-Oman/libraqm/issues/95
+   */
+  for (int i = rq->text_len - 2; i >= 0;  --i)
+  {
+    if (rq->text_info[i].script == HB_SCRIPT_INHERITED ||
+        rq->text_info[i].script == HB_SCRIPT_COMMON)
+      rq->text_info[i].script = rq->text_info[i + 1].script;
+  }
+
 #ifdef RAQM_TESTING
   RAQM_TEST ("After script detection:\n");
   for (size_t i = 0; i < rq->text_len; ++i)
@@ -1507,6 +1565,14 @@ _raqm_resolve_scripts (raqm_t *rq)
 static bool
 _raqm_shape (raqm_t *rq)
 {
+  hb_buffer_flags_t hb_buffer_flags = HB_BUFFER_FLAG_BOT | HB_BUFFER_FLAG_EOT;
+
+#if defined(HAVE_DECL_HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES) && \
+    HAVE_DECL_HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
+  if (rq->invisible_glyph < 0)
+    hb_buffer_flags |= HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES;
+#endif
+
   for (raqm_run_t *run = rq->runs; run != NULL; run = run->next)
   {
     run->buffer = hb_buffer_create ();
@@ -1516,7 +1582,13 @@ _raqm_shape (raqm_t *rq)
     hb_buffer_set_script (run->buffer, run->script);
     hb_buffer_set_language (run->buffer, rq->text_info[run->pos].lang);
     hb_buffer_set_direction (run->buffer, run->direction);
-    hb_buffer_set_flags (run->buffer, HB_BUFFER_FLAG_BOT | HB_BUFFER_FLAG_EOT);
+    hb_buffer_set_flags (run->buffer, hb_buffer_flags);
+
+#ifdef HAVE_HB_BUFFER_SET_INVISIBLE_GLYPH
+    if (rq->invisible_glyph > 0)
+      hb_buffer_set_invisible_glyph (run->buffer, rq->invisible_glyph);
+#endif
+
     hb_shape_full (run->font, run->buffer, rq->features, rq->features_len,
                    NULL);
   }
@@ -1530,7 +1602,7 @@ _raqm_u32_to_u8_index (raqm_t   *rq,
                        uint32_t  index)
 {
   FriBidiStrIndex length;
-  char *output = malloc ((sizeof (uint32_t) * index) + 1);
+  char *output = malloc ((sizeof (char) * 4 * index) + 1);
 
   length = fribidi_unicode_to_charset (FRIBIDI_CHAR_SET_UTF8,
                                        rq->text,
@@ -1794,11 +1866,10 @@ _raqm_allowed_grapheme_boundary (hb_codepoint_t l_char,
   hb_unicode_general_category_t l_category;
   hb_unicode_general_category_t r_category;
   _raqm_grapheme_t l_grapheme, r_grapheme;
+  hb_unicode_funcs_t* unicode_funcs = hb_unicode_funcs_get_default ();
 
-  l_category = hb_unicode_general_category (hb_unicode_funcs_get_default (),
-                                            l_char);
-  r_category = hb_unicode_general_category (hb_unicode_funcs_get_default (),
-                                            r_char);
+  l_category = hb_unicode_general_category (unicode_funcs, l_char);
+  r_category = hb_unicode_general_category (unicode_funcs, r_char);
   l_grapheme = _raqm_get_grapheme_break (l_char, l_category);
   r_grapheme = _raqm_get_grapheme_break (r_char, r_category);
 
@@ -1903,3 +1974,107 @@ _raqm_in_hangul_syllable (hb_codepoint_t ch)
   (void)ch;
   return false;
 }
+
+/**
+ * raqm_version:
+ * @major: (out): Library major version component.
+ * @minor: (out): Library minor version component.
+ * @micro: (out): Library micro version component.
+ *
+ * Returns library version as three integer components.
+ *
+ * Since: 0.7
+ **/
+void
+raqm_version (unsigned int *major,
+              unsigned int *minor,
+              unsigned int *micro)
+{
+  *major = RAQM_VERSION_MAJOR;
+  *minor = RAQM_VERSION_MINOR;
+  *micro = RAQM_VERSION_MICRO;
+}
+
+/**
+ * raqm_version_string:
+ *
+ * Returns library version as a string with three components.
+ *
+ * Return value: library version string.
+ *
+ * Since: 0.7
+ **/
+const char *
+raqm_version_string (void)
+{
+  return RAQM_VERSION_STRING;
+}
+
+/**
+ * raqm_version_atleast:
+ * @major: Library major version component.
+ * @minor: Library minor version component.
+ * @micro: Library micro version component.
+ *
+ * Checks if library version is less than or equal the specified version.
+ *
+ * Return value:
+ * %true if library version is less than or equal the specfied version, %false
+ * otherwise.
+ *
+ * Since: 0.7
+ **/
+bool
+raqm_version_atleast (unsigned int major,
+                      unsigned int minor,
+                      unsigned int micro)
+{
+  return RAQM_VERSION_ATLEAST (major, minor, micro);
+}
+
+/**
+ * RAQM_VERSION_ATLEAST:
+ * @major: Library major version component.
+ * @minor: Library minor version component.
+ * @micro: Library micro version component.
+ *
+ * Checks if library version is less than or equal the specified version.
+ *
+ * Return value:
+ * %true if library version is less than or equal the specfied version, %false
+ * otherwise.
+ *
+ * Since: 0.7
+ **/
+
+/**
+ * RAQM_VERSION_STRING:
+ *
+ * Library version as a string with three components.
+ *
+ * Since: 0.7
+ **/
+
+/**
+ * RAQM_VERSION_MAJOR:
+ *
+ * Library major version component.
+ *
+ * Since: 0.7
+ **/
+
+/**
+ * RAQM_VERSION_MINOR:
+ *
+ * Library minor version component.
+ *
+ * Since: 0.7
+ **/
+
+/**
+ * RAQM_VERSION_MICRO:
+ *
+ * Library micro version component.
+ *
+ * Since: 0.7
+ **/
